/*
 * File: printd.c
 *
 * Description:
 *
 * Author: Stefania Alborghetti - INFN Milan
 * e-mail: stefania.alborghetti@mi.infn.it
 *
 * History revision:
 * 25-Mar-2004 (FP) Modified flawed 'itostr' function. Funny it took so
 *                  long to notice.
 *  4-Aug-2009 (FP) Added support for IPv6 via getaddrinfo().
 *  9-Sep-2010 (FP) Moved str_ctr here (for cups_backend).
 */

#include "print.h"

#include <poll.h>

#ifdef VMS
#ifdef MULTINET
int vms_async_connect(int fd, struct sockaddr *addr,
                      int len, int timeout);
#endif
#endif

char* spool_first_line=NULL;
int _prints_socket_timeout = 45;
int _prints_lpc_status_requested=FALSE;

/*
 * str_print(FILE *fp, int sockfd, char *request, struct control *param)
 */


void str_print(FILE* fp, int sockfd, char *request, char *format, struct control *param, int silent_flag)

{
      int n;
      char  sendline[MAXLINE+1], recvline[MAXLINE+1];
      char buf[MAXLINE+1];
      char *file=NULL;
      int minusone = -1;

      sendline[0]='\0';
      recvline[0]='\0';
      buf[0]='\0';

      bzero(sendline,MAXLINE+1);
      bzero(recvline,MAXLINE+1);
      bzero(buf,MAXLINE+1);

      /* Write request to server (lpr, lpq o lprm) */
       if((wr_sock(request,sockfd)) !=0)
          err_sys("str_print:wr_sock error");

      file=param->file;
#if 0
      /*Eliminate PATH from file_name (if present at all) */
#ifdef VMS /*VMS style*/
      if((file=strrchr(param->file,']')) !=NULL){
         file++;
      }else {
         file=param->file;
      }
      for(n=0;n<=strlen(file);n++)
         if(file[n]==';') {
            file[n]='\000';
            break;
         }
#else /*UNIX style*/
      if((file=strrchr(param->file,'/')) !=NULL){
         file++;
      }else {
         file=param->file;
      }
#endif /*VMS*/
#endif

      if((wr_sock(file,sockfd))!=0)
          err_sys("str_print:wr_sock error");
      if((wr_sock(param->user,sockfd))!=0)
          err_sys("str_print:wr_sock error");
      if((wr_sock(param->host,sockfd))!=0)
          err_sys("str_print:wr_sock error");
      if((wr_sock(param->printer,sockfd))!=0)
          err_sys("str_print:wr_sock error");

     if ((n=wri_sock(sockfd,param->id_only)) == -1)
          err_sys("str_print: write error");

      if ((n=wri_sock(sockfd,param->uid)) == -1)
          err_sys("str_print: write error");

      if ((n=wri_sock(sockfd,param->reserve)) == -1)
          err_sys("str_print: write error");

      if ((n=wri_sock(sockfd,param->copies)) == -1)
          err_sys("str_print: write error");

      if((wr_sock(param->entry,sockfd))!=0)
          err_sys("str_print:wr_sock error");
      /*printf("Format is <%s>\n",format);*/
      if((wr_sock(format,sockfd))!=0)
          err_sys("str_print:wr_sock error");


      if(spool_first_line != NULL){
         n=strlen(spool_first_line);

         if ((wri_sock(sockfd,n)) == -1)
             err_sys("str_print: write error");
#if (defined(VMS))&&(defined(MULTINET))
          if ((socket_write(sockfd, spool_first_line, n)) != n)
             err_sys("str_print: socket_write error");
#else
          if ((write(sockfd, spool_first_line, n)) != n)
             err_sys("str_print: write error");
#endif
      free(spool_first_line);
      }

      while ((n=fread(sendline,sizeof(char),MAXLINE,fp)) > 0) {
          if ((wri_sock(sockfd,n)) == -1)
             err_sys("str_print: write error");
#if (defined(VMS))&&(defined(MULTINET))
          if ((socket_write(sockfd, sendline, n)) != n)
             err_sys("str_print: socket_write error");
#else
          if ((write(sockfd, sendline, n)) != n)
             err_sys("str_print: write error");
#endif
      }
      if ((wri_sock(sockfd,minusone)) == -1)
        err_sys("str_print: write error");
/*readline returns zero if there is nothing left to read or if '\001' has
  been received*/
      while( (n = readline(sockfd, recvline, MAXLINE)) != 0) {
          if (n < 0 )
             err_sys("str_cli: readline error");
          recvline[n] ='\000';
          if (silent_flag == FALSE) fputs(recvline, stdout);
      }

}  

/*
 *  int wr_sock(char* name, int sockfd);                            *
 */
int wr_sock(char *name,int sockfd)
{
    int n;
    char *temp_string;

    temp_string = (char *)malloc(strlen(name)+2);
    sprintf(temp_string,"%s\n",name);
    n = strlen(temp_string);
#if (defined(VMS))&&(defined(MULTINET))
    if (socket_write(sockfd, temp_string, n) != n)
     {
      free(temp_string);
      err_sys("wr_sock: socket_write error");
     }
#else
    if (write(sockfd, temp_string, n) != n)
     {
      free(temp_string);
      err_sys("wr_sock: socket write error");
     }
#endif
    free(temp_string);
    return(0);
}

/*
 *  int wri_sock (int sockfd, int value);                            *
 *  writes integer to socket - performs htonl conversion
 */
int wri_sock(int sockfd,int value)
{
    int net_value;
    net_value=htonl(value);
#if (defined(VMS))&&(defined(MULTINET))
    if (socket_write(sockfd, &net_value, sizeof(int)) != sizeof(int))
        err_sys("wri_sock: socket_write error");
#else
    if (write(sockfd, &net_value, sizeof(int)) != sizeof(int))
        err_sys("wri_sock: socket write error");
#endif
    return(0);
}

int readline(int fd, char *ptar, int maxlen)
{
    int rc;
    char    c;
    char* ptr; 
    struct pollfd read_poll;
    int pollret;
    
    ptr=ptar;
    
    if (maxlen<=1) return(-1);
    
    *ptr = 0;

    while ( 1 ) {
#if (defined(VMS))&&(defined(MULTINET))
       if ( (rc = socket_read(fd, &c, 1)) != 1) break;
#else
       rc = -1;
       read_poll.fd = fd;
       read_poll.events = ( POLLIN | POLLERR | POLLHUP | POLLNVAL );
       pollret = poll(&read_poll, 1, _prints_socket_timeout*1000);
       if (pollret <= 0) break; /* poll error or timeout */
       if ((read_poll.revents & POLLIN) == 0) break; /* Other anomalous events */
       if ( (rc = read(fd, &c, 1)) != 1) break;
#endif
       if (ptr<(ptar+maxlen-1)) *ptr++ = c;
	  
       if (c == '\n') break;
       
       if (c == '\001') return 0;
    }
       
    if (rc < 0) {
       *ptr=0;
        return(-1);
    }
     
    *ptr = 0;
    return(ptr-ptar);
}


int readint(int fd, int *ptar)
{
    int readl;
    char *readp; 
    int  n;
    struct pollfd read_poll;
    int pollret;
    
    *ptar = 0;

    readp=(char *)ptar;
    readl=sizeof(int);
    while ( 1 ) {
#if (defined(VMS))&&(defined(MULTINET))
      if ((n=socket_read(fd,readp,readl)) <= 0 ) break;
#else
 
      n = -1;
      read_poll.fd = fd;
      read_poll.events = ( POLLIN | POLLERR | POLLHUP | POLLNVAL );
      pollret = poll(&read_poll, 1, _prints_socket_timeout*1000);
      if (pollret <= 0) break; /* poll error or timeout */
      if ((read_poll.revents & POLLIN) == 0) break; /* Other anomalous events */
      if ((n=read(fd,readp,readl)) <= 0) break;
#endif
      readp+=n;
      readl-=n;
      if(readl<=0) break;
    }

    if (n < 0) {
       *ptar=0;
        return(-1);
    }

   *ptar = ntohl(*ptar);
        
   return(sizeof(int));
}

int writen(int fd, char *ptr, int nbytes)
{
    int nleft, nwritten;
    nleft = nbytes;
    while (nleft > 0) {
       nwritten = write(fd,ptr,nleft);
       if (nwritten <= 0)
           return(nwritten);    
       nleft -= nwritten;
       ptr += nwritten;
    }
    return (nbytes - nleft);
} 

/*
 * pid_t root(char *file_root)
 */

root(char *file_root)
{
    int pid,n;
    char buf[8];
    buf[0]='\0';
    bzero(buf,8);
    file_root[0]='\0';
    bzero(file_root,MAXWORD);
    /*Ottiene proprio pid */
    pid=getpid();  
    /* printf("The pid is: %i \n", pid); */
    sprintf(buf,"%i",pid);
    n=strlen(buf);
    strcat(file_root,"temp.");
    strncat(file_root,buf,n);
    /* fputs(file_root,stdout); */
    return pid;
}   

char* itostr(unsigned int i)
{
 char *retbuf;
 int n_digits;

 n_digits = (int)i/10 + 1;
 retbuf = (char *)malloc(n_digits + 2);
 if (retbuf == NULL) return NULL;
 sprintf(retbuf,"%u\000",i);
 return retbuf;
}


/*
 * int chop(char* str1, int size)
 */

int chop(char* str1, int size)  
{ 
   int n;
   char *temp;

   temp=(char *)malloc(sizeof(char));

   n=strlen(str1);
   temp=(char *)realloc(temp,n);
   bzero(temp,size);
   strncpy(temp,str1,n-1);
   bzero(str1,size);
   strncpy(str1,temp,n-1);                                       
   free(temp);
   return 1; 
}   

/******************************************************************************       
 *  int is_alive(char* server)
 ******************************************************************************/

is_alive(char* server)
{
   int sockfd;

   sockfd = sock_init(server);

   if (sockfd >= 0) {
#if (defined(VMS))&&(defined(MULTINET))
      socket_close(sockfd);
#else
      close(sockfd);
#endif
      return 1;
   }

   return 0;
}

/***************************************************************************************
 *  int  does_resolve(char* server)                                               *
 ***************************************************************************************/

int
does_resolve(char* server)
{
    struct addrinfo ai_req, *ai_ans;

    ai_req.ai_flags = 0;
    ai_req.ai_family = PF_UNSPEC;
    ai_req.ai_socktype = SOCK_STREAM;
    ai_req.ai_protocol = 0; /* Any protocol is OK */
                            /* Any port is OK */
    if (getaddrinfo(server, NULL, &ai_req, &ai_ans) != 0) {
      return 0;
    }

    freeaddrinfo(ai_ans);
    return 1;
}

/***************************************************************************************
 *  int  sock_init(char* server)                                               *
 ***************************************************************************************/
      
int
sock_init(char* server)
{
    int    sockfd,is_al;
    int    status;
    static char *last_resolution=NULL;
    static struct addrinfo *ai_ans=NULL, *last_ans=NULL;
    struct addrinfo ai_req, *cur_ans;
    char port_string[16];
    fd_set writefs;
    struct timeval timeout; 
    int retcod;
           
    /* No secondary server */
    if (server==NULL) return -1;
    if (strlen(server)==0) return -1;

    /* Get destination address */
  
    if (last_resolution == NULL || (strcmp(last_resolution, server) != 0)) {
       if (last_resolution != NULL) {
          free(last_resolution);
          freeaddrinfo(ai_ans);
          last_resolution = NULL;
          ai_ans = NULL;
       }
       ai_req.ai_flags = 0;
       ai_req.ai_family = PF_UNSPEC;
       ai_req.ai_socktype = SOCK_STREAM;
       ai_req.ai_protocol = 0; /* Any protocol is OK */
       snprintf(port_string,sizeof(port_string),"%5d",SERV_TCP_PORT);
       if ((retcod = getaddrinfo(server, port_string, &ai_req, &ai_ans)) != 0)
          err_sys("server: getaddrinfo returns error");
    }
    if (last_resolution != NULL) {
       cur_ans = last_ans;
    }
    else cur_ans = ai_ans;

    /* Search for a working transport */
    for(;cur_ans != NULL; cur_ans=cur_ans->ai_next) {
       if ( (sockfd = socket (cur_ans->ai_family, cur_ans->ai_socktype,
                              cur_ans->ai_protocol)) < 0) continue;

#ifndef VMS    
       if ((status=fcntl(sockfd,F_GETFL,NULL)) == -1) {
          close(sockfd);
          continue;
       }
       status |= FNDELAY;  
       if ((fcntl(sockfd,F_SETFL,status)) == -1) {
          close(sockfd);
          continue;
       }
       connect (sockfd, cur_ans->ai_addr, cur_ans->ai_addrlen); 
#else
#ifdef MULTINET
       vms_async_connect(sockfd, cur_ans->ai_addr, cur_ans->ai_addrlen,1);
#else
       connect (sockfd, cur_ans->ai_addr, cur_ans->ai_addrlen); 
#endif /*MULTINET*/
#endif /*VMS*/

       FD_ZERO(&writefs);
       FD_SET(sockfd,&writefs);
       timeout.tv_sec=TIMEOUT_CLI; 
       timeout.tv_usec=0; 
       is_al=select(FD_SETSIZE,(fd_set *) 0,&writefs,(fd_set *) 0,&timeout); 
       if (is_al <= 0) {

#if (defined(VMS))&&(defined(MULTINET))
          socket_close(sockfd);
#else
          close(sockfd);
#endif
          continue;
       }   

#if (defined(VMS))&&(defined(MULTINET))
       socket_close(sockfd);
#else
       close(sockfd);
#endif

       if ( (sockfd = socket (cur_ans->ai_family, cur_ans->ai_socktype,
                              cur_ans->ai_protocol)) < 0) continue;
       if ((status=connect (sockfd, cur_ans->ai_addr, cur_ans->ai_addrlen)) == -1){
#if (defined(VMS))&&(defined(MULTINET))
          socket_close(sockfd);
#else
          close(sockfd);
#endif
	  continue;
       } else {
           /* Everything's OK. Let's save the contact info. */
           last_resolution = strdup(server);
           if (last_resolution != NULL) {
              last_ans = cur_ans;
           }
	   return sockfd;
       }    	     
   }
    
   freeaddrinfo(ai_ans);
   ai_ans = NULL;
   last_resolution = NULL;

   return -1;       
}   

/***********************************************************************************
 * int enquire(char* server)
 ***********************************************************************************/

int enquire(char* server)
{
    int    n, sockfd;
    char read_line[MAXWORD];
    
    /* No secondary server */
    if (strlen(server)==0) return 0;

    sockfd=sock_init(server);       
    write(sockfd,"n_jobs\n",7);  
    /* Shitcan version number from other server */
    if(readline(sockfd, read_line, MAXWORD)<=0)
    {
       fprintf(stderr,"enquire:readline failed");   
       close(sockfd);
       return -1;
    }
    if ((readint(sockfd,&n)) < 0)
    {
       fprintf(stderr,"enquire:readint (for number of jobs) failed");   
       close(sockfd);
       return -1;
    }
    close(sockfd);
    return(n);   
}   

/*
 *   _prints_check_data_file(param.file)
 *   Return 0 if file is postscript
 *   Return 1 if file is PDF
 *   Return 2 if file is ASCII
 */
 
int  _prints_check_data_file(char *file,int status)
{
  FILE* fd;
  char c1,c2,c3,c4;
  int _prints_stdin=FALSE;
  
  
  if(status ==0) /*Called by client*/
   if ((strstr(file,"stdin"))!=NULL){ 
     _prints_stdin=TRUE;
     fd=stdin;
   } 
   
   if(_prints_stdin==FALSE)
    {
     if ((fd=fopen(file,"r")) == NULL)
       err_sys("_prints_check_data_file: Cannot open file");
   
     c1 = fgetc(fd);
     c2 = fgetc(fd);
     c3 = fgetc(fd);
     c4 = fgetc(fd);
   
     fclose(fd);
    }
   else
    {
     c1 = fgetc(fd);
     c2 = fgetc(fd);
     c3 = fgetc(fd);
     c4 = fgetc(fd);
     if (c4!=EOF)
      if (ungetc(c4,stdin)<0)
       fprintf(stderr,"Cannot push <%c> back into stdin:%s\n",c4,strerror(errno)); 
     if (c3!=EOF)
      if (ungetc(c3,stdin)<0)
       fprintf(stderr,"Cannot push <%c> back into stdin:%s\n",c3,strerror(errno)); 
     if (c2!=EOF)
      if (ungetc(c2,stdin)<0)
       fprintf(stderr,"Cannot push <%c> back into stdin:%s\n",c2,strerror(errno)); 
     if (c1!=EOF)
      if (ungetc(c1,stdin)<0)
       fprintf(stderr,"Cannot push <%c> back into stdin:%s\n",c1,strerror(errno)); 
    }
/*Check first 4 characters of file*/   
   
   if(c1=='%' && c2=='!') return 0; /*File is postscript*/
   else if(c1=='%' && c2=='P' &&
      c3=='D' && c4=='F') return 1; /*File is PDF*/  
   else return 2; /*File is ASCII*/
}     

/*
 * str_ctr
 */


str_ctr(int sockfd, char *request, struct control *param, int process_output)

{
      int n;
      char recvline[MAXLINE + 1];
      char *message;
      int _prints_lpr=FALSE;
      int _prints_lpq=FALSE;
      int _prints_lprm=FALSE;
      int _prints_lpc=FALSE;
      char *pipchr;
      
      if((wr_sock(request,sockfd))!=0)
          err_sys("print client: error writing to socket");
       
      if(strcmp(request,"lpq")== 0) _prints_lpq=TRUE;
      else if(strcmp(request,"lpr")== 0) _prints_lpr=TRUE;
      else if(strcmp(request,"lprm")==0) _prints_lprm=TRUE;
      else if(strcmp(request,"lpc")==0) _prints_lpc=TRUE;
      else err_sys("print client: error in request type"); 
      
      if((_prints_lpq) || (_prints_lpr))
          if((wr_sock(param->file,sockfd))!=0)
          err_sys("print client: error writing to socket");
      if(_prints_lprm) 
        if((wr_sock(param->entry,sockfd))!=0)
          err_sys("print client: error writing to socket");   
      
      if((wr_sock(param->user,sockfd))!=0)
          err_sys("print client: error writing to socket");
      if((wr_sock(param->host,sockfd))!=0)
          err_sys("print client: error writing to socket");
      if((wr_sock(param->printer,sockfd))!=0)
          err_sys("print client: error writing to socket"); 
      
      
      if ((n=wri_sock(sockfd,param->id_only)) == -1)
          err_sys("print client: error writing to socket");
      
      if ((n=wri_sock(sockfd,param->uid)) == -1)
          err_sys("print client: error writing to socket");

      if(_prints_lpr) {
        if ((n=wri_sock(sockfd,param->reserve)) == -1)
           err_sys("print client: error writing to socket"); 	  
	
        if ((n=wri_sock(sockfd,param->copies)) == -1)
           err_sys("print client: error writing to socket"); 	  
        
	if((wr_sock(param->entry,sockfd))!=0)
          err_sys("print client: error writing to socket");
      }

      if(_prints_lpc) {
        if ((n=wri_sock(sockfd,param->reserve)) == -1)
           err_sys("print client: error writing to socket");

        if (param->lpc_message == NULL) message = "";
        else                           message=param->lpc_message;
        if((wr_sock(message,sockfd))!=0)
          err_sys("print client: error writing to socket");
        if (param->lpc_message!=NULL) free(param->lpc_message);
        param->lpc_message = NULL;
      }
      
     if (process_output) {
         /* Read server responses. Write them to stdout and exit. */
         while( (n = readline(sockfd, recvline, MAXLINE)) != 0) {
            if (n < 0 )
              err_dump("print client: error reading from socket");
            recvline[n] =0;
            if (!_prints_lpc_status_requested)
             {
                 fputs(recvline, stdout);
             } else {
                 /* Process lpq output and format status. */
                 if ((pipchr = strchr(recvline,'|')) != NULL)
                  {
                      *pipchr = '\000'; /* Truncate line to printer name. */
                      printf("%s:\n\tqueuing is enabled\n\tprinting is enabled\n",
                             recvline);
                  }
             }
         }
     }
}     
