/*
  File :     common.h

  Author :   Stefania Alborghetti ($Author: root $)
  e-mail :   "stefania.alborghetti@mi.infn.it"

  Revision history :
   9-Oct-1997 Original release
   8-Feb-1999 Added macros for "reserve" flag.
  25-Mar-2004 (FP) Added 'itostr' prototype.
   9-Sep-2010 (FP) Moved str_ctr here (for cups_backend).

  Description: Common macros and prototypes for the print server
               and client.
*/

#ifndef PrintmiCOMMONh
#define PrintmiCOMMONh

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <fcntl.h>
#include        <errno.h>
#include        <unistd.h>
#include        <pwd.h>

#ifdef __hpux
#define FILENAME_MAX 256
#endif

#if defined(__sun) && defined(__sparc)
#include <sys/file.h>
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

#ifndef VMS
#include        <sys/select.h>
/*#include        <malloc.h>*/
#include        <sys/types.h>
#include        <sys/time.h> 
/*#include        <sys/mode.h>*/
#include        <sys/stat.h>
#include        <sys/wait.h>
#include        <sys/socket.h>
#include        <netinet/in.h>
#include        <arpa/inet.h>
#else
#ifdef MULTINET
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]time.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include.arpa]inet.h"
#include "multinet_root:[multinet.include]netdb.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#else
#include        <types.h>
#include        <socket.h>
#include        <in.h>
#include        <inet.h>
#include        <time.h>
#endif /*MULTINET*/
#include        <stat.h>
#endif /*VMS*/

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define   SERV_TCP_PORT      5168
#define MAXLINE 512
#define MAXWORD 80
#define MAXFORMAT 60
#define MAXENTRY 20
#define REQ_SIZE 10
#define IP_ADDR_SIZE 16
#define N_SERVER 11
#define PSPRINT_DB "./psprintcap"
#define SERV_DB "./server_list"
#define ERROR -1
#define TIMEOUT 2
#define TIMEOUT_CLI 5
#define TIMEOUT_SERV 5

extern char **environ;
extern int _prints_debug;
extern int _prints_socket_timeout;
extern int _prints_lpc_status_requested;

#define PRINTMI_READLINE_BUFFER_LENGTH 1024

#ifdef VERSION
#define SERVER_VERSION VERSION
#define CLIENT_VERSION "0.1.2a"
#else
#define SERVER_VERSION "3.0"
#define CLIENT_VERSION "0.1.2a"
#endif /*VERSION*/

struct control{
  char file[FILENAME_MAX];
  char user[MAXENTRY];
  char host[MAXENTRY];
  char printer[MAXENTRY];
  char entry[MAXENTRY];
  char *lpc_message;
  unsigned int copies;
  uid_t uid;
  int reserve;
  int id_only;
};  

typedef struct control control;

void str_print(FILE* fp, int sockfd, char *request, char *format, struct control *param, int silent_flag);
int str_ctr(int sockfd, char *request, struct control *param, int process_output);
int wr_sock(char *name,int sockfd);
int wri_sock (int sockfd, int value);

int does_resolve(char *server);
int is_alive(char *server);

int sock_init(char*);
int readline(int fd, char *ptar, int maxlen);

int err_sys(char*, ...);
int err_dump(char*, ...);
void err_ret(char*, ...);

char* itostr(unsigned int i);

#define PRINTMI_RESERVE_NOP 0
#define PRINTMI_RESERVE_RES 1
#define PRINTMI_LPC_DISABLE 1
#define PRINTMI_LPC_ENABLE 0
#define PRINTMI_MOVE_REQ 100

#define DSPRINTF_INCREASE_SIZE 1016
#define DSPRINTF_MAX_SIZE 8192*DSPRINTF_INCREASE_SIZE
#define DSTRING int *
#define DSTRING_PTR(dst) ((char*)(dst+2))

#endif /* def PrintmiCOMMONh */
