/*
 * File: error.c
 *
 * Description:
 *
 * Author: Stefania Alborghetti - INFN Milan
 * e-mail: stefania.alborghetti@mi.infn.it
 *
 * History revision:
 *  1-Feb-1999 (FP) Initial entry.
 * 31-May-2004 (FP) Moved to stdarg.h, as varargs seems not to
 *                  be supported anymore as of gcc 3.3.3.
 *
 *  Description.
 *  Error handling routine. Can be used for a client or server by
 *  defining CLIENT or SERVER. CLIENT is the default.
 */

#include   <stdio.h>
#include   <string.h>
#include   <stdlib.h>
#if (defined(VMS))&&(defined(MULTINET))
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include]errno.h"
#else
#include   <errno.h>
#endif
#include   <stdarg.h>
/*#include   "systype.h" */
#define BSD

#ifdef CLIENT
#ifdef SERVER
cant define both CLIENT and SERVER
#endif
#endif
#ifndef CLIENT
#ifndef SERVER
#define CLIENT 1
#endif
#endif
#ifndef NULL
#define NULL  ((void *) 0)
#endif

int my_perror();

char *pname = NULL;

#ifdef CLIENT   /* Everything goes to stderr */


/* Fatal error. Print message and exit.
 * err_quit(str, arg1,arg2, ...) 
 * str indicates the required format for any argument.
 */

/*VARARGS1*/
   
int
err_quit(char *fmt,...)
{
    va_list   args;
    if (pname != NULL)
       fprintf(stderr, "%s: ", pname);
    va_start(args, fmt);
    vfprintf(stderr, fmt,args);
    va_end(args);
    fputc('\n', stderr);
    exit(1);
}     




/* Errore fatale. Stampa un messaggio e termina.
 * Non esegue il dump del core ma stampa il valore di errno di sistema
 * ed il suo messaggio associato
 * err_sys(str, arg1,arg2, ...) 
 * La stringa str deve specificare la conversione per un qualsiasi arg.
 */

/*VARARGS1*/
  
int
err_sys(char *fmt,...)
{
    va_list   args;
    if (pname != NULL)
       fprintf(stderr, "%s: ", pname);
    va_start(args,fmt);
    vfprintf(stderr, fmt,args);
    va_end(args);
    my_perror();
    /*sig_handler();*/
    exit(1); 
}    




/* Errore recuperabile. Stampa un messaggio e ritorna al chiamante.
 * err_ret(str, arg1,arg2, ...) 
 * La stringa str deve specificare la conversione per un qualsiasi arg.
 */

/*VARARGS1*/
void 
err_ret(char *fmt,...)
{
    va_list   args;
    if (pname != NULL)
       fprintf(stderr, "%s: ", pname);
    va_start(args, fmt);
    vfprintf(stderr, fmt,args);
    va_end(args);
    my_perror();
    fflush(stdout);
    fflush(stderr);
    return;
}   




/* Errore fatale. Fa il dump del core e termina.
 * err_dump(str, arg1,arg2, ...) 
 * La stringa str deve specificare la conversione per un qualsiasi arg.
 */

/*VARARGS1*/
   
int
err_dump(char *fmt,...)
{
    va_list   args;
    if (pname != NULL)
       fprintf(stderr, "%s: ", pname);
    va_start(args, fmt);
    vfprintf(stderr, fmt,args);
    va_end(args);
    my_perror();
    fflush(stdout);
    fflush(stderr);
    abort();   /* esegue il dump e termina */
    exit(1);   /* non dovrebbe arrivare qui */
}    



/* Stampa il valore errno di Unix  */

int
my_perror()
{
    char *sys_err_str();
#if (defined(VMS))&&(defined(MULTINET))
    fprintf(stderr, " %s %s\n", vms_errno_string(), sys_err_str());
#else
    fprintf(stderr, "%s\n", sys_err_str());
#endif
}

#endif /* CLIENT */

#define MAX_REASONABLE_ERRNO 151

char *sys_err_str()
{  
    static char msgstr[200];
    if (errno != 0) {
        if (errno > 0 && errno < MAX_REASONABLE_ERRNO) 
           sprintf(msgstr, "(%s)", strerror(errno));
        else
           sprintf(msgstr, "(errno = %d)", errno);
    } else {
        msgstr[0] = '\0';
    }
   return(msgstr);
}        
