#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#ifndef PRINTMI_BIN_AREA
#define PRINTMI_BIN_AREA "/usr/local/printmi"
#endif
#ifndef PRINTIF_FILTER_NAME
#define PRINTIF_FILTER_NAME "printif"
#endif

int main(int argc, char *argv[])
 {

  int pipe_from_child[2];
  char buffer[256];
  char lpq_name[FILENAME_MAX];
  pid_t child_process,retpid;
  int read_length;
  int status;
  char *printer="all";
  FILE *fout = stdout;
  FILE *sout;
  int i;
  char readc, *div, *desc;

  if (argc>1 && (strncmp(argv[1],"-h",2)==0))
   {
    fprintf(stderr,"Usage: %s [<output file>] [<printer queue name>]\n",
            argv[0]);
    exit(1);
   }

  if (argc>2) printer = argv[2];
  if (argc>2)
   {
    if ((fout = fopen(argv[1],"w"))==NULL)
     {
      fprintf(stderr,"%s: Cannot open file <%s>: %s\n",argv[0],argv[1],strerror(errno));
      exit(1);
     }
   }

  if (pipe(pipe_from_child)<0)
   {
    fprintf(stderr,"%s: Cannot create pipe: %s\n",argv[0],strerror(errno));
    exit(1);
   }

  for (i=0; i<2; i++)
   {
    if ((status=fcntl(pipe_from_child[i],F_GETFL,NULL)) != -1)
     {
      status |= FNDELAY;
      fcntl(pipe_from_child[i],F_SETFL,status);
     }
   }

  if ((child_process = fork())<0)
   {
    fprintf(stderr,"%s: Cannot fork child: %s\n",argv[0],strerror(errno));
    exit(1);
   }

  sprintf(lpq_name,"%s/lpq",PRINTMI_BIN_AREA);

  if (child_process == 0)
   {
    dup2(pipe_from_child[1],1);

    execlp(lpq_name,"lpq","-j",printer,NULL);
   }

  wait(&status);

  for (;;)
   {
    i = 0;
    readc = 0;
    while ((i<(sizeof(buffer)-1)) && (readc != '\n'))
     {
      read_length = read(pipe_from_child[0],&readc,1);
      if (read_length < 0)
       {
        if (errno != EAGAIN)
          fprintf(stderr,"%s: Error reading pipe: %s\n",argv[0],strerror(errno));
        close(pipe_from_child[0]);
        close(pipe_from_child[1]);
        fclose(fout);
        exit(0);
       }
      if (read_length==1) buffer[i++] = readc;
      if (read_length==0) break;
     }
    buffer[i-1]='\000';
    if ((div = strchr(buffer,'|'))!= NULL)
     {
      desc = div + 1;
      *div = '\000';
      fprintf(fout,"%s:%s:\\\n",buffer,desc);
      fprintf(fout,"        :af=%s:mx#0:\\\n",buffer);
      fprintf(fout,"        :lp=/dev/null:sh:if=%s/%s:sd=/var/spool/lpd/from_printmi:\n",
              PRINTMI_BIN_AREA,PRINTIF_FILTER_NAME);
     }
    if (read_length==0) break;
   }

  close(pipe_from_child[0]);
  close(pipe_from_child[1]);
  exit(0);

 }

