/*
  File :     lpd_client.h


  Author :   Francesco Prelz ($Author: printmi $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
   9-Oct-1997 Original release
  10-Feb-1998 Added LPD_CONNECT_TIMEOUT
  14-Sep-1998 Added LPD_WAIT_FOR_SOCKET, LPD_CLIENT_WATCHDOG_TIMEOUT,
              lpd_client_hard_timeout()
  22-Sep-1999 Increased watchdog timeout to 40 minutes.

  Description: Common macros and prototypes for the library
               in lpd_client.c
*/

#ifndef LpdClientH
#define LpdClientH

#define LPD_SERVER_NAME "lpd.mi.infn.it"
#define LPD_TALKER_PORT_MAX 731
#define LPD_TALKER_PORT_MIN 721

                            /*seconds*/
#define LPD_STATUS_TIMEOUT 3
#define LPD_CONNECT_TIMEOUT 3
#define LPD_WAIT_FOR_SOCKET 3
/* This in case a given printer seizes the LPD connection and doesn't give it
   back. Shutdown the connection and remove the job after 40 minutes */
#define LPD_CLIENT_WATCHDOG_TIMEOUT 2400

#define LPD_POSTSCRIPT_TYPE 1
#define LPD_IS(mask,file_type) ((file_type)&(mask))

#ifndef SHUT_RDWR
#define SHUT_RDWR 2
#endif

int get_lpd_connection(char *lpd_server);
int read_lpd_ack(int fd);
int read_lpd_response(int fd, char *buffer, unsigned int size); 
char *compose_lpd_control_file(char *host, char *user, char *name,
                               char *serv_name, int file_type, int banner);
int lpd_client_spool_file(char *lpd_server, char *queue,
                          char *host, char *user, char *orig_name,
                          char *name, int file_type, int queue_no, int banner);
int lpd_client_get_queue_state(char *lpd_server, char *queue,
                               char *buffer, unsigned int size);
int lpd_client_remove_job(char *lpd_server, char *queue,
                               char *user, unsigned int entry);

void lpd_client_hard_timeout();

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#endif /*LpdQueueH*/
