/*
 * File: print.c
 *
 * Description:  
 *
 * Author: Stefania Alborghetti - INFN Milan
 * e-mail: stefania.alborghetti@mi.infn.it
 *
 * History revision:
 * 13-Mar-1998 (FP) Added name lookup for servers.
 * 25-Mar-1998 (FP) Moved str_print, wr_sock, wri_sock,
 *                  _prints_get_primary_server,_prints_get_backup_server,
 *                  to common.c (in order to share with printif.c)
 * 15-Jul-1998 (FP) Added option to print a page range (invokes psselect
 *                  in the server).
 * 13-Aug-1998 (FP) Added VMS system service calls for pre-v7 systems.
 *  1-Feb-1999 (MC+FP) Implemented VMS qualifier functionalities.
 *                     + Full messages in VMS STYLE implemented.
 *  8-Feb-1999 (FP) Implemented job move (-M) flag.
 * 30-Apr-1999 (FP) Added option to print multiple files.
 * 31-Jan-2008 (FP) Added lpc 'status' command to make Java happy.
 *  4-Aug-2009 (FP) Ported to IPv6. Removed static name resolution
 *                  to go along with changes in sock_init (common.c).
 * 28-May-2010 (FP) Removed buffer overflow in 'gethostname' call
 *                  that went undetected since the origin of this code.
 *  9-Sep-2010 (FP) Moved str_ctr to common.c (for cups_backend)
 * 29-Sep-2010 (FP) gethostname can  now return an error if the
 *                  supplied buffer is too small. Act accordingly.
 *
 */


#include   "print.h"

#ifdef VMS
#include <starlet.h>
#include <jpidef.h>
typedef struct _ile3 {
#pragma __nomember_alignment
    unsigned short int ile3$w_length;   /* Length of buffer in bytes        */
    unsigned short int ile3$w_code;     /* Item code value                  */
    void *ile3$ps_bufaddr;              /* Buffer address                   */
    unsigned short int *ile3$ps_retlen_addr; /* Address of word for returned length */
    } ile3;

#ifdef WANT_VMS_CLD_FORMAT
/* ------------------------------------------------------------------ */
/* 	Massimo Carboni extention for OpenVMS system, using .CLD      */
/* ------------------------------------------------------------------ */
 
#include <clidef.h>
#include <climsgdef.h>
#include <descrip.h>
#include <rmsdef.h>
#endif
#endif

/*Global variables*/
int _prints_reserved_or_existing_job=FALSE;
int _prints_silent=FALSE;
int _prints_say_version=FALSE;
char* format=NULL;

/*External variables*/
extern char* spool_first_line;

static struct control param;

int main(int argc, char *argv[])
{
    int sockfd;
    FILE* fp;
    struct passwd* _prints_user_pwd;
    
    char request[REQ_SIZE];
    char* user, *printer;
    char *pname;
    int lprm_from_secondary=FALSE;
    int i;
    int extra_files,extra_argc;
#ifdef VMS
    unsigned int iosb[2];
    unsigned short vms_length;
    ile3 vms_sysreq;
#endif
    
    format=(char*) malloc(sizeof(char));
    *format='\000';

#ifdef VMS
    vms_sysreq.ile3$w_length = sizeof(param.uid);
    vms_sysreq.ile3$w_code = JPI$_UIC;
    vms_sysreq.ile3$ps_bufaddr = (void *)&(param.uid);
    vms_sysreq.ile3$ps_retlen_addr = &vms_length;
    sys$getjpiw(0, 0, 0, &vms_sysreq, iosb, 0, 0);

    vms_sysreq.ile3$w_length = sizeof(param.user);
    vms_sysreq.ile3$w_code = JPI$_USERNAME;
    vms_sysreq.ile3$ps_bufaddr = (void *)param.user;
    vms_sysreq.ile3$ps_retlen_addr = &vms_length;
    sys$getjpiw(0, 0, 0, &vms_sysreq, iosb, 0, 0);
    if (vms_length<sizeof(param.user)) param.user[vms_length] = '\000';
    /* Trim trailing spaces */
    for (i=strlen(param.user)-1;i>=0;i--)
     {
      if (param.user[i]==' ' || param.user[i]=='\t') param.user[i]='\000';
      else break;
     }
#else
    param.uid=getuid();
    if((_prints_user_pwd=getpwuid(param.uid)) == NULL)
       err_sys("client:getpwuid failed");
    user=_prints_user_pwd->pw_name;
    strncpy(param.user,user,sizeof(param.user));
#endif
    
    if((gethostname(param.host,sizeof(param.host))) == -1)
     {
#ifdef ENAMETOOLONG
       if (errno != EINVAL && errno != ENAMETOOLONG)  
#else
       if (errno != EINVAL)  
#endif
          err_sys("Client: gethostname failed");
     }
    /* gethostname can leave truncated names behind */
    param.host[sizeof(param.host)-1] = '\000';
	
    if((printer=getenv("PRINTER")) == NULL) param.printer[0]='\000';
    else strncpy(param.printer,printer,sizeof(param.printer));    	
    
    param.reserve=PRINTMI_RESERVE_NOP;
    param.id_only=0;
    param.copies=1;
    param.lpc_message = NULL;
    *param.file='\000';
    *param.entry='\000';
    

#ifdef VMS
    argv++;
    argc--;
#ifdef WANT_VMS_CLD_FORMAT
    pname = strtok((char *)argv[0], "/");
#else
    pname = argv[0];
#endif
#else
    pname = argv[0];
#endif
    
    
    if(strstr(pname,"lp") == (pname + strlen(pname) - 2)){
      strncpy(request,"lpr",sizeof(request)); 
      request[sizeof(request)-1]='\000'; 
      if((_prints_lp_get_opt(argc,argv,&extra_files,&extra_argc))!=0)
         err_sys("lp: Incorrect set of options.");
    }  
    else if(strstr(pname,"lpr") == (pname + strlen(pname) - 3)){
      strncpy(request,"lpr",sizeof(request));
      request[sizeof(request)-1]='\000';
      if((_prints_lpr_get_opt(argc,argv,&extra_files,&extra_argc)) !=0)
        err_sys("lpr: Incorrect set of options.");
    }  
    else if(strstr(pname,"lpq") == (pname + strlen(pname) - 3)){
      strncpy(request,"lpq",sizeof(request));
      request[sizeof(request)-1]='\000';
      if((_prints_lpq_get_opt(argc,argv)) !=0)
        err_sys("lpq: Incorrect set of options.");
    } 
    else if(strstr(pname,"lpc") == (pname + strlen(pname) - 3)){
      if((_prints_lpc_get_opt(argc,argv)) !=0)
        err_sys("lpc: Incorrect set of options.");
      if (!_prints_lpc_status_requested) 
        strncpy(request,"lpc",sizeof(request));
      else
        strncpy(request,"lpq",sizeof(request));
      request[sizeof(request)-1]='\000';
    }
    else if(strstr(pname,"lprm") == (pname + strlen(pname) - 4)){
      strncpy(request,"lprm",sizeof(request));
      request[sizeof(request)-1]='\000'; 
      if(argc == 2) strncpy(param.entry,argv[1],sizeof(param.entry));
      else if ((argc==3)&&(strstr(argv[1],"-S")!=NULL)) {
        strncpy(param.entry,argv[2],sizeof(param.entry));
        lprm_from_secondary = TRUE;
      }
      else {
        printf("Usage: %s [-S] entry\n",pname);
	exit(-1);
      }	
    } 
    else if(strstr(pname,"cancel") == (pname + strlen(pname) - 6)){
      if(argc == 2) {
         strncpy(request,"lprm",sizeof(request));   
         request[sizeof(request)-1]='\000';
         strncpy(param.entry,argv[1],sizeof(param.entry));
      }
      else {
        printf("Usage: %s entry\n",pname);
	exit(-1);
      }	
    } 

    
    if (lprm_from_secondary) { 
       /* Explicit request to remove job from the secondary server*/
       if ((is_alive(SERV_HOST_NAME_II)) == 0)  /*Primary serv. down*/
          err_sys("client: no response from backup print server");
       sockfd=sock_init(SERV_HOST_NAME_II);   
    } else {
       if ((is_alive(SERV_HOST_NAME_I)) == 0) { /*Primary serv. down*/
          if ((is_alive(SERV_HOST_NAME_II)) == 0)  /*Secondary serv. down*/
             err_sys("client: no response from print server(s)");
          sockfd=sock_init(SERV_HOST_NAME_II);   
       } else {
           sockfd=sock_init(SERV_HOST_NAME_I);
       }
    }                 


    if(_prints_say_version == TRUE)
      display_version(sockfd);

    /*Let's check if we are a valid client version*/
    if((_prints_check_version(sockfd))== -1){
       printf("Sorry: print client and server versions are incompatible.\n");
       printf("Please contact your system administrator.\n");
       exit(0);
    }   
    
    if(strcmp(request,"lpr") == 0 ) {    
       if(_prints_reserved_or_existing_job == TRUE ) str_ctr(sockfd,request,&param,TRUE);
       else { 
         if ((strcmp(param.file,"(stdin)") == 0)||
	      (strcmp(param.file,"-") == 0)) {
	         strncpy(param.file,"stdin",sizeof(param.file)); 
		 param.file[sizeof(param.file)-1]='\000';
                 fp = stdin;
	 }	 
         else if ((fp = fopen(param.file, "r")) == NULL )
           err_sys("client: cannot open input file");
#ifdef WANT_VMS_STYLE_MESSAGES
	 printf("%%LPR-I-VMSVER, PRINTMI VMS port.\n");
	 printf("%%LPR-I-PRINT, ");
#endif
         for (i=0; i<=extra_files; i++)
          {
           str_print(fp, sockfd, request, format, &param, _prints_silent);
           if ((extra_argc + i) > 0 && i < extra_files)
            {
             strncpy(param.file, argv[extra_argc + i], sizeof(param.file));
             fclose(fp);
             if ((fp = fopen(param.file, "r")) == NULL )
               err_sys("client: cannot open input file");
             close(sockfd);
             if ((is_alive(SERV_HOST_NAME_I)) == 0) { /*Primary serv. down*/
              if ((is_alive(SERV_HOST_NAME_II)) == 0)  /*Secondary serv. down*/
                 err_sys("client: no response from print server(s)");
              sockfd=sock_init(SERV_HOST_NAME_II);   
             } else {
                sockfd=sock_init(SERV_HOST_NAME_I);
             }
              if((_prints_check_version(sockfd))== -1){
                close(sockfd);
                fclose(fp);
                exit(0);
              }
            }
          }
         fclose(fp);
       }	 
       exit(0);
    }else if (strcmp(request,"lpq") ==0 ) {
#ifdef WANT_VMS_STYLE_MESSAGES
      printf("%%LPQ-I-VMSVER, PRINTMI VMS port.\n");
      printf("%%LPQ-I-QUEUE, ");
#endif
      str_ctr(sockfd, request, &param, TRUE);
      exit(0);
    }else if (strcmp(request,"lpc") ==0 ) {
       str_ctr(sockfd, request, &param, TRUE);
       exit(0);
    }else if (strcmp(request,"lprm") ==0 ) {
#ifdef WANT_VMS_STYLE_MESSAGES
      printf("%%LPRM-I-VMSVER, PRINTMI VMS port\n");
      printf("%%LPRM-I-INFO, ");
#endif 
       str_ctr(sockfd, request, &param, TRUE); 
       exit(0);
    } else {
         err_sys("print:Undefined option.\n");
    } 
#if (defined(VMS))&&(defined(MULTINET))
    socket_close(sockfd);
#else
    close(sockfd);      
#endif
    free(format);
    exit(0);
}       

/*
 * int _prints_check_version(int sockfd);
 */
 
int _prints_check_version(int sockfd)
{
   int n;
   char read_line[MAXLINE+1];
   char *serv_vers;
   
   if((n = readline(sockfd, read_line, MAXLINE)) <=0 ) 
       err_sys("print client: readline error");
    
   if((n >= 1)&&(n<=MAXLINE)) read_line[n-1]='\000';
   serv_vers = strchr(read_line, '/');
   if (serv_vers!=NULL) serv_vers+=2;
    
    if ((strncmp(CLIENT_VERSION,read_line,strlen(CLIENT_VERSION)))!=0){
        printf("Client version is: %s.\n",CLIENT_VERSION);
        printf("Server version is: %s.\n",serv_vers);
        printf("Warning: client was released for server version %s.\n",
                SERVER_VERSION);
        return(-1);
    }	       
    return(0);
} 


/*
 * int display_version(int sockfd);
 */
 
#ifdef VMS /*VMS complains about non void fun. not containing a return statem.*/
display_version(sockfd)
#else
int display_version(sockfd)
#endif /*VMS*/
{ 
  int n=0;
  char read_line[MAXLINE+1];
  char *serv_vers;

/* Write request to server: vers */
    if((wr_sock("vers",sockfd)) !=0)
       err_sys("display_version:wr_sock error");

    if((wr_sock(param.host,sockfd))!=0)
       err_sys("display_version:wr_sock error");

    if((n = readline(sockfd, read_line, MAXLINE)) <=0 ) 
       err_sys("display_version: readline error");
    
    if((n > 1)&&(n<=MAXLINE)) read_line[n-1]='\000';
    serv_vers = strchr(read_line, '/');
    if (serv_vers!=NULL) 
     {
      *serv_vers = '\000';
      serv_vers+=2;
     }
    
    printf("Client version is: %s.\n",CLIENT_VERSION);
    printf("Server version is: %s.\n",serv_vers);
    
    if ((strstr(read_line,CLIENT_VERSION))==NULL)
       printf("Warning: server was released for client version %s.\n",
             read_line);
	       
    exit(0);
}

/*
 *   _prints_lpr_get_opt(int argc, char** argv)
 */

int _prints_lpr_get_opt(int argc,char** argv, int *extra_files, int *extra_argc)
{
 int i,option;
 char *opt_str="u:y:P:T:W:#:nN:O:K:rR:M:p:B:D:e:f:s:a:d:l:m:t:cjqV";
 int found_file = FALSE;
 int _prints_numbering=FALSE;
 int _prints_set_font_scale=FALSE;
 int _prints_set_font=FALSE;
 int _prints_set_title=FALSE;
 int _prints_set_paper=FALSE;
 int _prints_set_scale=FALSE;
 int _prints_set_line_width=FALSE;
 int _prints_set_lines_per_page=FALSE;
 int _prints_set_margin=FALSE;
 int _prints_set_boxes=FALSE;
 int _prints_set_column=FALSE;
 int _prints_set_width=FALSE;
 int _prints_no_format=FALSE;
 int _prints_no_input_file=FALSE;
 char* _prints_width=NULL;
 char* _prints_paper=NULL;
 char* _prints_scale=NULL;
 char* _prints_line_width=NULL;
 char* _prints_lines_per_page=NULL;
 char* _prints_margin=NULL;
 char* _font_type=NULL;
 char* _font_scale=NULL;
 char* _prints_title=NULL;
 char* _prints_boxes=NULL;
 char* _prints_sides=NULL;
 char* _prints_border=NULL;
 char* _prints_copies=NULL;
 struct stat file_stat;
 char *sys_err_str();
 

 *extra_files = 0;
 *extra_argc = 0;

#if (defined(VMS)&&defined(WANT_VMS_CLD_FORMAT))
 
  /* ------------------------------------------------------------------ */
  /* 	Massimo Carboni extention for OpenVMS system, using .CLD        */
  /* ------------------------------------------------------------------ */

  register status; 
  short len;
  char value[80];
  $DESCRIPTOR (value_desc, value);
  $DESCRIPTOR (line,      "FILE");
  $DESCRIPTOR (queue,     "QUEUE");
  $DESCRIPTOR (psize,     "PAPERSIZE");
  $DESCRIPTOR (border,    "BORDER");
  $DESCRIPTOR (column,    "COLUMN");
  $DESCRIPTOR (sfactor,   "SCALEFACTOR");
  $DESCRIPTOR (font,      "FONT");
  $DESCRIPTOR (rjobid,    "RJOBID");
  $DESCRIPTOR (sides,     "SIDES");
  $DESCRIPTOR (wborder,   "WBORDER");
  $DESCRIPTOR (fformat,   "FORMAT");
  $DESCRIPTOR (boxes,     "BOXES");
  $DESCRIPTOR (landscape, "LANDSCAPE");
  $DESCRIPTOR (portrait,  "PORTRAIT");
  $DESCRIPTOR (numbering, "NUMBERING");
  $DESCRIPTOR (tray,      "TRAY");
  $DESCRIPTOR (fontsize,  "FSIZE");
  $DESCRIPTOR (lines,     "LINESPERPAGE");
  $DESCRIPTOR (tumble,    "TUMBLE");
  $DESCRIPTOR (title,     "TITLE");
  $DESCRIPTOR (width,     "WIDTH");
  $DESCRIPTOR (version,   "VERSION");
  $DESCRIPTOR (range,     "RANGE");
  $DESCRIPTOR (ncopies,   "COPIES");
  $DESCRIPTOR (reserv,    "RESERV");
  $DESCRIPTOR (jobres,    "JOBRESERV");

  /* ------------------------------------------------------------------ */
  /* 	Get file to print  param.file                                   */
  /* ------------------------------------------------------------------ */

  if (CLI$PRESENT (&line) == CLI$_ABSENT)
    printf("Parameter FILE is absent.\n");
  else if (!((status = CLI$GET_VALUE (&line, &value_desc, &len)) & 1)) 
    lib$stop(status);
  else {
    value[len] = 0;
    strncpy(param.file,value,sizeof(param.file));
  }

  /* ------------------------------------------------------------------ */
  /* QUEUE name ---> param.printer                                      */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&queue) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&queue, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      strncpy(param.printer,value,sizeof(param.printer));
    }
  }

  /* ------------------------------------------------------------------ */
  /* SIDES name ---> param.printer                                      */
  /* Grupped options SIDES and TUMBLE: SIDES=2,SIDES=TUMBLE             */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&sides) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&sides, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      if ( strcmp(value,"TUMBLE") == 0 ) {
	format=(char*) realloc(format,strlen(format)+7);
        strcat(format," -ton");
      }
      format=(char*) realloc(format,strlen(format)+strlen(value)+4); 
      strcat(format," -K");	  
      _prints_sides="2";
      strcat(format,_prints_sides);
    }
  }

  /* ------------------------------------------------------------------ */
  /* PAPERSIZE                                                          */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&psize) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&psize, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_paper=TRUE;	      
      _prints_paper=value;
    }
  }

  /* ------------------------------------------------------------------ */
  /* BORDER                                                             */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&border) != CLI$_ABSENT &&
      CLI$PRESENT (&border) != CLI$_NEGATED) {
      format=(char*) realloc(format,strlen(format)+strlen(value)+4);    
      strcat(format," -B");
      _prints_border="y";      
      strcat(format,_prints_border);
  }

  /* ------------------------------------------------------------------ */
  /* NUMBERING                                                          */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&numbering) != CLI$_ABSENT &&
      CLI$PRESENT (&numbering) != CLI$_NEGATED) {
      format=(char*) realloc(format,strlen(format)+strlen(value)+4);    
      _prints_numbering=TRUE;      
  }

  /* ------------------------------------------------------------------ */
  /* COLUMN                                                             */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&column) != CLI$_ABSENT) &&
      CLI$PRESENT (&column) != CLI$_NEGATED) {
    if (CLI$GET_VALUE (&column, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_column=TRUE;
    }
  }


  /* ------------------------------------------------------------------ */
  /* SCALEFACTOR                                                        */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&sfactor) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&sfactor, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_scale=TRUE;	      
      _prints_scale=value;
    }
  }


  /* ------------------------------------------------------------------ */
  /*   FONT                                                                 */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&font) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&font, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_font=TRUE;
      _font_type=value;
    }
  }


  /* ------------------------------------------------------------------ */
  /* RJOBID                                                             */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&rjobid) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&rjobid, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      param.id_only=1;
    }
  }


  /* ------------------------------------------------------------------ */
  /* WBORDER                                                            */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&wborder) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&wborder, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_line_width=TRUE;      
      if(value!=NULL) {	      
         _prints_line_width=value; 
      }

    }
  }

  /* ------------------------------------------------------------------ */
  /* BOXES                                                              */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&boxes) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&boxes, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_set_boxes=TRUE;
      _prints_boxes=value;
    }
  }


  /* ------------------------------------------------------------------ */
  /* COPIES                                                             */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&ncopies) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&ncopies, &value_desc, &len) != CLI$_ABSENT) {
      value[len] = 0;
      _prints_copies=value;  
      if(_prints_copies != NULL) param.copies=atoi(_prints_copies);
      if (param.copies > MAX_COPY_NUMBER)
         printf ("Warning: %d copies sent to spooling server!\n",
	                                              param.copies);    
      if (param.copies == 0) {
         printf("Number of copies is equal to zero\n");
         exit(-1);
      }
    }
  }


  /* ------------------------------------------------------------------ */
  /* VERSION                                                            */
  /* ------------------------------------------------------------------ */
  if (CLI$PRESENT (&version) != CLI$_ABSENT) {
    if (CLI$GET_VALUE (&version, &value_desc, &len) == CLI$_ABSENT){
      _prints_say_version=TRUE;
    } else { 
      value[len] = 0;
      _prints_say_version=TRUE;
    }
  }


  /* ------------------------------------------------------------------ */
  /*                                                                    */
  /* ------------------------------------------------------------------ */
  /*   if (CLI$PRESENT (&) != CLI$_ABSENT) { */
  /*     printf("Qualifier  is present "); */
  /*     if (CLI$GET_VALUE (&, &value_desc, &len) == CLI$_ABSENT) */
  /*       printf("does not have a value.\n"); */
  /*     else{  */
  /*       value[len] = 0; */
  /*       printf("value is: (%s)\n", value); */
  /*     } */
  /*   } */

  /* ------------------------------------------------------------------ */
  /*                                                                    */
  /* ------------------------------------------------------------------ */
  /*   if (CLI$PRESENT (&) != CLI$_ABSENT) { */
  /*     printf("Qualifier  is present "); */
  /*     if (CLI$GET_VALUE (&, &value_desc, &len) == CLI$_ABSENT) */
  /*       printf("does not have a value.\n"); */
  /*     else{  */
  /*       value[len] = 0; */
  /*       printf("value is: (%s)\n", value); */
  /*     } */
  /*   } */


#else 

 while ((option = getopt(argc, argv, opt_str)) != -1) {
  switch(option){ 
    case 'a':
      _prints_set_paper=TRUE;	      
      _prints_paper=optarg;
      break; 
    case 'B':
      format=(char*) realloc(format,strlen(format)+strlen(optarg)+4);    
      strcat(format," -B");
      _prints_border=optarg;      
      strcat(format,_prints_border);
      break;
    case 'c':
      _prints_set_column=TRUE;
      break;          
    case 'D': /*psprint script may pass data type but we ignore it (so far)*/
      break; 
    case 'd':
      _prints_set_scale=TRUE;	      
      _prints_scale=optarg;
      break;
    case 'e':
      format=(char*) realloc(format,strlen(format)+strlen(optarg)+4);
      strcat(format," -e");
      strcat(format,optarg);
      break;  
    case 'f':
      _prints_set_font=TRUE;
      _font_type=optarg;
      break;
    case 'j':         	 		 	  
      param.id_only=1;
      break;   
    case 'K':
      format=(char*) realloc(format,strlen(format)+strlen(optarg)+4); 
      strcat(format," -K");	  
      _prints_sides=optarg;        
      strcat(format,_prints_sides);
      break;  
    case 'l':
      _prints_set_line_width=TRUE;      
      if(optarg!=NULL) {	      
         _prints_line_width=optarg; 
      }
      else _prints_line_width="1";
      break;
    case 'm':
      _prints_set_margin=TRUE;	      
      _prints_margin=optarg;
      break;                
    case 'N':
      _prints_set_boxes=TRUE;
      _prints_boxes=optarg;
      break; 
    case 'n':
      _prints_no_format=TRUE;
      break;  
    case 'O':
      format=(char*) realloc(format,strlen(format)+4);
      strcat(format," ");
      if((strstr(optarg,"landscape")) != 0)
         strcat(format,"-l");
      else if((strstr(optarg,"portrait")) != 0)
         strcat(format,"-n");
      else {
         fprintf(stderr,"%s: Unrecognized orientation mode: %s\n",
	         argv[0],optarg);
         exit(-1);
      }
      break;
    case 'P':
      strncpy(param.printer,optarg,sizeof(param.printer));
      break;
    case 'p':
      if((strstr(optarg,"y")) != NULL)
        _prints_numbering=TRUE;
      break; 
    case 'q':
      _prints_silent=TRUE;
      break;   
    case's':
      _prints_set_font_scale=TRUE;	      
      _font_scale=optarg;
      break;
    case't':
      if((strstr(optarg,"off"))!=NULL){
        format=(char*) realloc(format,strlen(format)+8);
        strcat(format," -toff");
      } else if ((strstr(optarg,"on"))!=NULL){ 
        format=(char*) realloc(format,strlen(format)+7);
        strcat(format," -ton");
      }
      break;  
    case'T':
      _prints_set_title=TRUE;	      
      _prints_title=optarg;
      break;       
    case 'u':
      _prints_set_lines_per_page=TRUE;
      if(optarg!=NULL) {
         _prints_lines_per_page=optarg;
      }
      else _prints_lines_per_page="25";
      break;
    case 'W':
      _prints_set_width=TRUE;	      
      _prints_width=optarg;
      break;
    case'#':	
      _prints_copies=optarg;  
      if(_prints_copies != NULL) param.copies=atoi(_prints_copies);
      if (param.copies > MAX_COPY_NUMBER)
         printf ("Warning: %d copies sent to spooling server!\n",
	                                              param.copies);    
      if (param.copies == 0) {
         printf("Number of copies is equal to zero\n");
         exit(-1);
      }
      break; 
    case 'r':         	 		 	  
      param.reserve=PRINTMI_RESERVE_RES;
      break;
    case 'R':
      strncpy(param.entry,optarg,sizeof(param.entry)); 
      _prints_reserved_or_existing_job=TRUE; 
      break;
    case 'M':
      strncpy(param.entry,optarg,sizeof(param.entry)); 
      param.reserve=PRINTMI_MOVE_REQ;
      _prints_reserved_or_existing_job=TRUE; 
      break;
    case 'V':
      _prints_say_version=TRUE;
      break;
    case 'y':
      format=(char*) realloc(format,strlen(format)+strlen(optarg)+6);
      for (i=0;i<strlen(optarg);i++)
        if (optarg[i]<'0' && optarg[i]>'9' && optarg[i]!='-' && optarg[i]!=',')
         break;
      if (i<strlen(optarg))
       {
        fprintf(stderr,"Warning: page range must contain only numbers, dashes and commas. Ignored.\n");
       }
      else
       {
        strcat(format," -y");
        strcat(format,optarg);
       }
      break;
    default:
      _prints_lpr_usage(argv[0]);
   }   
 } 
#endif /*def VMS&&WANT_VMS_CLD_FORMAT*/

 if (optind < argc && argc > 1) {
     found_file = TRUE;
#if (defined(VMS)&&defined(WANT_VMS_CLD_FORMAT))
     strncpy(param.file,strtok((char *)argv[optind], "/"),sizeof(param.file));
#else
     strncpy(param.file,argv[optind],sizeof(param.file));
#endif
     if((stat(param.file,&file_stat))!=0)
#ifdef WANT_VMS_STYLE_MESSAGES
       {
	 printf("%%LPR-W-NOFILES, file <%s> not found \n",param.file);
	 exit(-1);
       }
#else
       if((strstr(sys_err_str(),"No such file or directory"))!=NULL){
       printf("lpr: couldn't find file %s.\n",param.file);
       exit(-1);
       } else err_sys("lpr:stat failed");
#endif  
     if(file_stat.st_size <= 0){
#ifdef WANT_VMS_STYLE_MESSAGES
       printf("%%LPR-W-FEMPTY, file is empty\n");
#else
       printf("lpr: %s is an empty file.\n",param.file);
#endif
       exit(-1);
     }
     if (argc - optind > 1) {
     *extra_files = (argc - optind -1);
     *extra_argc = optind + 1;
    }
 } else if ((optind == argc) || (argc == 1)){
        strncpy(param.file,"(stdin)",sizeof(param.file));
	param.file[sizeof(param.file)-1]='\000';
   }	
 
   if(_prints_no_format==TRUE) {
       format=(char*) realloc(format,strlen(format)+4);   
       strcat(format," -x");
   }	  
 
   if((_prints_reserved_or_existing_job == TRUE)||(_prints_say_version==TRUE))
       _prints_no_input_file=TRUE;
   
   if(_prints_no_input_file!= TRUE) {

    if (strlen(param.printer)==0)
     {
#ifdef WANT_VMS_STYLE_MESSAGES
       printf ("%%LPR-E-QUEUE, no queue specified\n");
       printf ("%%LPR-I-USAGE, LPR/QUEUE=<> file.ext\n");
       exit(-1);
#else
      fprintf(stderr,
       "%s: no printer specified (-P) and the PRINTER env. variable is not set.\n",
       argv[0]);
      _prints_lpr_usage(argv[0]);
#endif
     }

    if((_prints_check_data_file(param.file,0)) == 0){ /*File is postscript*/
       if(_prints_set_column==TRUE) {
          format=(char*) realloc(format,strlen(format)+4);
          strcat(format," -c");
       }	  
       if(_prints_set_boxes == TRUE){
          format=(char*) realloc(format,strlen(format)+strlen(_prints_boxes)+4);
	  strcat(format," -N");
          strcat(format,_prints_boxes);
       }
       if(_prints_numbering == TRUE){  
	 printf("Numbering is not allowed with PostScript files.\n");
         printf("Ignoring option -p\n");
       } 
       if(_prints_set_font == TRUE){ 
          printf("Selecting font is not allowed with PostScript files.\n");
          printf("Ignoring option -f %s\n",_font_type);
       }
       if(_prints_set_font_scale == TRUE){
          printf("Selecting font size is not allowed with PostScript files.\n");
          printf("Ignoring option -s %s\n",_font_scale);
       }
       if(_prints_set_title == TRUE){
          printf("Setting a title is not allowed with PostScript files.\n");
          printf("Ignoring option -T %s\n",_prints_title);
       }
       if(_prints_set_width==TRUE){
          printf("Setting width is not allowed with PostScript files.\n");
	  printf("Ignoring option -W %s\n",_prints_width);
       }	  
       if(_prints_set_lines_per_page == TRUE){
          printf("Setting lines per page is not allowed with PostScript files.\n");
          printf("Ignoring option -u %s\n",_prints_lines_per_page);
       }
       if(_prints_set_paper == TRUE){
        format=(char*) realloc(format,strlen(format)+strlen(_prints_paper)+4);
        strcat(format," -P");
	strcat(format,_prints_paper);
       }
       if(_prints_set_scale == TRUE){
        format=(char*) realloc(format,strlen(format)+strlen(_prints_scale)+4);
        strcat(format," -d");
	strcat(format,_prints_scale);
       }
       if(_prints_set_line_width == TRUE){
        format=(char*) realloc(format,strlen(format)
	                        +strlen(_prints_line_width)+4);
	strcat(format," -b");
	strcat(format,_prints_line_width);
       }
       if(_prints_set_margin == TRUE){
        format=(char*) realloc(format,strlen(format)
	                          +strlen(_prints_margin)+4);
        strcat(format," -m");
	strcat(format,_prints_margin);
       }	
     }else { /*File is ASCII*/
       if(_prints_set_column==TRUE){
         printf("Selecting column order is not allowed with text files.\n");
         printf("Ignoring option -c\n");
       }
       if(_prints_set_boxes == TRUE){
          if((strstr(_prints_boxes,"8")==0)&&(strstr(_prints_boxes,"4")==0)&&
            (strstr(_prints_boxes,"2")==0)&&(strstr(_prints_boxes,"1")==0)){
            fprintf(stderr,"Boxes per page for ASCII files: 8 4 2 1\n");
            exit(-1);
          }
	  format=(char*) realloc(format,strlen(format)
	                               +strlen(_prints_boxes)+4);
          strcat(format," -N");
          strcat(format,_prints_boxes);
       }
       if(_prints_numbering == TRUE){
          format=(char*) realloc(format,strlen(format)+4);
          strcat(format," -p");
       }	   
       if(_prints_set_font == TRUE){ 
          format=(char*) realloc(format,strlen(format)+strlen(_font_type)+4);
          strcat(format," -f");
          strcat(format,_font_type);
	}
	if(_prints_set_font_scale == TRUE){
	   format=(char*) realloc(format,strlen(format)+strlen(_font_scale)+4);
	   strcat(format," -z");
           strcat(format,_font_scale);
	} 
	if(_prints_set_title == TRUE){
	   format=(char*) realloc(format,strlen(format)
	                           +strlen(_prints_title)+4);
	   strcat(format," -h");
           strcat(format,_prints_title);
	}   
       if(_prints_set_width==TRUE){
         format=(char*) realloc(format,strlen(format)
	                          +strlen(_prints_width)+4);
	 strcat(format," -W");
         strcat(format,_prints_width);
       }
       if(_prints_set_lines_per_page == TRUE){
        format=(char*) realloc(format,strlen(format)
                                +strlen(_prints_lines_per_page)+4);
        strcat(format," -u");
        strcat(format,_prints_lines_per_page);
       }
       if(_prints_set_paper == TRUE){
        format=(char*) realloc(format,strlen(format)+strlen(_prints_paper)+4);
        strcat(format," -P");
	strcat(format,_prints_paper);
       }
       if(_prints_set_scale == TRUE){
        printf("Selecting scale is not allowed with text files.\n");
        printf("Ignoring option -d %s\n",_prints_scale);
       }
       if(_prints_set_line_width == TRUE){
        printf("Selecting border width is not allowed with text files.\n");
        printf("Ignoring option -l %s\n",_prints_line_width);
       }
       if(_prints_set_margin == TRUE){
        printf("Selecting margin is not allowed with text files.\n");
        printf("Ignoring option -m %s\n",_prints_margin);
       }     
      }	
    }   
    
 /*printf("Format in get_lpr_opt is <%s>\n",format);*/
 return 0;
}  
            

/*
 *   _prints_lp_get_opt(int argc, char** argv)
 */


int _prints_lp_get_opt(int argc,char** argv,int *extra_files, int *extra_argc)
{
 int option;
 char *opt_str="c:d:n:rR:t:jqV";
 int found_file = FALSE;
 int _prints_set_title=FALSE;
 int _prints_no_input_file=FALSE;
 struct stat file_stat;
 char *sys_err_str();
 char* _prints_title=NULL;
 char* _prints_copies=NULL;
 
 while ((option = getopt(argc, argv, opt_str)) != -1) {
  switch(option){       
    case 'c':
      _prints_copies=optarg;
      if(_prints_copies != NULL) param.copies=atoi(_prints_copies);
      if (param.copies > MAX_COPY_NUMBER)
         printf ("Warning: %d copies sent to spooling server!\n",
	                                              param.copies);
      if (param.copies == 0) {
         printf("Number of copies is equal to zero\n");
         exit(-1);
      }
      break;    
    case 'd':
      strncpy(param.printer,optarg,sizeof(param.printer));
      break;
    case 'j':         	 		 	  
      param.id_only=1;
      break; 
    case 'q':
      _prints_silent=TRUE;
      break;   
    case't':
      _prints_set_title=TRUE;	      
      _prints_title=optarg;
      break;       
    case'n':	
      _prints_copies=optarg;
      if(_prints_copies != NULL) param.copies=atoi(_prints_copies);
      if (param.copies > MAX_COPY_NUMBER)
         printf ("Warning: %d copies sent to spooling server!\n",
	                                              param.copies);
      if (param.copies == 0) {
         printf("Number of copies is equal to zero\n");
         exit(-1);
      }
      break;    
    case 'r':         	 		 	  
      param.reserve=PRINTMI_RESERVE_RES;
      break;
    case 'R':
      strncpy(param.entry,optarg,sizeof(param.entry)); 
      _prints_reserved_or_existing_job=TRUE; 
      break;
   case 'V':
      _prints_say_version=TRUE;
      break;
    default:
      _prints_lp_usage(argv[0]);
   }   
 }
 
 *extra_files = 0;
 *extra_argc = 0;

 if (optind < argc && argc > 1) {
     found_file = TRUE;
     strncpy(param.file,argv[optind],sizeof(param.file));
     if((stat(param.file,&file_stat))!=0)
     if((strstr(sys_err_str(),"No such file or directory"))!=NULL){
       printf("lp: cannot access %s\n",param.file);
       exit(-1);
       } else err_sys("lp: stat failed"); 
     if(file_stat.st_size <= 0){
       printf("lp: %s is an empty file.\n",param.file);
       exit(-1);
     }  
     if (argc - optind > 1) {
      *extra_files = (argc - optind - 1);
      *extra_argc = optind + 1;
    }
 }                    
 if (!(found_file)) {
    strncpy(param.file,"(stdin)",sizeof(param.file));
    param.file[sizeof(param.file)-1]='\000';
 }   

 if((_prints_reserved_or_existing_job == TRUE)||(_prints_say_version== TRUE))
       _prints_no_input_file=TRUE;
   
  if(_prints_no_input_file!= TRUE) { 
   if((_prints_check_data_file(param.file,0)) == 0){
     if(_prints_set_title == TRUE){
          printf("Setting a title is not allowed with PostScript files.\n");
          printf("Ignoring option -T %s\n",_prints_title);
     }
     }else{      
     if(_prints_set_title == TRUE){
	   format=(char*) realloc(format,strlen(format)
	                             +strlen(_prints_title)+4);
	   strcat(format," -h");
           strcat(format,_prints_title);
     }
   }
 }          
 return 0;
}              

/*
 * _prints_lpq_get_opt(int argc, char** argv)
 */

int _prints_lpq_get_opt(int argc, char** argv)
{
  int option;
  char *opt_str="P:p:j";
  while ((option = getopt(argc, argv, opt_str)) != -1) {
   switch(option){ 
     case 'j':         	 		 	  
       param.id_only=1;
       break;
     case 'P':
       strncpy(param.printer,optarg,sizeof(param.printer)); 
       break;
     case 'p':
       strncpy(param.printer,optarg,sizeof(param.printer)); 
       break;
     default:
#ifdef WANT_VMS_STYLE_MESSAGES
       printf("%%LPQ-E-PRINTER, use LPQ printer\n");
#else
       printf("Usage: lpq printer or lpq -p printer or lpq -P printer\n");  
#endif
       exit(0);
   }
  }
  if (optind < argc && argc > 1) 
     strncpy(param.printer,argv[optind],sizeof(param.printer));                   

  return(0);
}  	      	  

/*
 * _prints_lpc_get_opt(int argc, char** argv)
 */

int _prints_lpc_get_opt(int argc, char** argv)
{
#define PRINT_LPC_USAGE "Usage: lpc status or lpc disable|down|enable printer [message]\n"

  int option;
  int message_length=0;
  int i;

  if (argc < 3)
   {
       if (strcmp(argv[1],"status") == 0)
        {
            _prints_lpc_status_requested = TRUE;
            strncpy(param.printer,"all",sizeof(param.printer));
            param.id_only=1;
            return(0);
        }

       fprintf(stderr,PRINT_LPC_USAGE);
       exit(1);
   }

  if (strcmp(argv[1],"status") == 0)
   {
       _prints_lpc_status_requested = TRUE;
       param.id_only=1;
       strncpy(param.printer,argv[2],sizeof(param.printer));
       return(0);
   }

  if (getuid() != 0)
   {
       fprintf(stderr,"Sorry. Non-superusers can only use lpc status.\n");
       exit(1);
   }
  if (strcmp(argv[1],"disable")!=0 && strcmp(argv[1],"enable")!=0 && 
      strcmp(argv[1],"down")!=0)
   {
       fprintf(stderr,PRINT_LPC_USAGE);
       exit(1);
   }

  if (strcmp(argv[1],"disable")==0 || strcmp(argv[1],"down")==0)
       param.reserve = PRINTMI_LPC_DISABLE;
  else param.reserve = PRINTMI_LPC_ENABLE;
  
  strncpy(param.printer,argv[2],sizeof(param.printer));

  if (argc>3)
   {
    for (i=3;i<argc;i++) message_length += (strlen(argv[i]) + 1);
    message_length++;
    param.lpc_message = (char *)realloc(param.lpc_message,message_length);
    if (param.lpc_message!=NULL)
     {
      *(param.lpc_message) = '\000';
      for (i=3;i<argc;i++)
       {
        strcat(param.lpc_message,argv[i]);
        strcat(param.lpc_message," ");
       }
     }
   }

  return(0);
}          

/*
 *  _prints_lpr_usage(char* pname);
 */
 
void _prints_lpr_usage(char* pname)
{
   printf("Usage:  %s [options] [-r|-R entry] [-M entry] [-P printer] file\n",
                   pname);
   printf("\tOptions are:\n");
   printf("\t\t-a paper_size\n");
   printf("\t\t-B y|n Border                (default is no)\n");
   printf("\t\t-c                           (only for POSTSCRIPT files)\n");
   printf("\t\t    (prints boxes in column order instead of in row order.)\n");
   printf("\t\t-d dimension_scale_factor    (only for POSTSCRIPT files)\n");
   printf("\t\t-f font_type                 (only for ASCII files)\n");
   printf("\t\t-j                           (return job ID only)\n");
   printf("\t\t-K sides\n");
   printf("\t\t-l border_width              (only for POSTSCRIPT files)\n");
   printf("\t\t-m margin_size               (only for POSTSCRIPT files)\n");
   printf("\t\t-n                           (disable any format option)\n");
   printf("\t\t-N boxes                     (for ASCII files: 1,2,4,8 boxes)\n");
   printf("\t\t                       (for POSTSCRIPT files any whole number.)\n");
   printf("\t\t-O landscape/portrait\n");
   printf("\t\t-p y|n numbering             (only for ASCII files)\n");
   printf("\t\t-e tray# | M                 (# of tray requested or Manual feed\n");
   printf("\t\t-s font_dimension            (only for ASCII files)\n");
   printf("\t\t-u lines_per_page            (only for ASCII files)\n");
   printf("\t\t-t on|off                    (tumble mode, default is on)\n");
   printf("\t\t-T title                     (only for ASCII files)\n");
   printf("\t\t                             (and with numbering on).\n");
   printf("\t\t-W width                     (only for ASCII files)\n");
   printf("\t\t-V                           (print out version and exit)\n");
   printf("\t\t-y range                     (print only a page range, e.g. 2-3)\n");
   printf("\t\t-# copies\n");
   exit(-1);
}   



/*
 *  _prints_lp_usage(char* pname);
 */
 
void _prints_lp_usage(char* pname)
{
   printf("Usage: %s [-r|-R entry] [-d printer] [-n copies | -c copies] file\n",pname);
   exit(-1);
}


