/*
  File :     printd.h

  Author :   Stefania Alborghetti ($Author: root $)
  e-mail :   "stefania.alborghetti@mi.infn.it"

  Revision history :
   9-Oct-1997 Original release
  23-Jan-2004 Changed exec_gs prototype to accommodate extra GS arguments.
              Added extra_gs_arguments prototype.
  25-Mar-2004 Removed itostr prototype (moved to common.h)

  Description: Macros and prototypes for the spooler
               queue daemon main program.
*/

#ifndef PrintmiPRINTDh
#define PrintmiPRINTDh

#include "common.h"
#ifdef OSF1
#include <sys/mode.h>
#include <sys/uswitch.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <time.h>

#define   PRINT_PROGRAM_ASCII    "/var/spool/printmi/mpage"
#define   PRINT_PROGRAM_POST     "/var/spool/printmi/psnup"
#define   PRINT_PROGRAM_PSSELECT "/var/spool/printmi/psselect"
#define   PRINTD_REGULAR_USER    "printmi"
#define   PRINT_GHOSTSCRIPT      "/usr/bin/gs" 
#define   PRINT_PDF_2_PS         "/usr/bin/pdf2ps" 

#ifndef SHUT_RDWR
#define SHUT_RDWR 2
#endif

/*Database containing queues names*/
#define SPOOL_SQUEUES_DB "queues.db"
#define SPOOL_ACCESS_CONFIG "hosts.printmi"
#define PRINTD_PID_FILE "printmi.pid"

#define MAX_STATUS_MESSAGE_LEN 200
#define PRINTS_MAX_REQUESTS 3

typedef struct allowed_host_s {
  char name[MAXHOSTNAMELEN];
  struct allowed_host_s *next;
} allowed_host;

extern int _prints_primary;
extern char *_prints_other_server;

DSTRING lpr_wrap(struct control* param, char* buf, pid_t pid);
DSTRING lpq_wrap(struct  control* param);
DSTRING lprm_wrap(struct control* param);
DSTRING lpc_wrap(struct control* param);

int is_duplex(char* queue_name);
char *use_gs(char *queue_name);
char **extra_gs_arguments(char *queue_name, int *n_arguments);
char *check_file(char *queue_name);
char *get_alias_switches(char *queue_name);
char *dealias_queue_name(char *queue_name);
int isempty_wrap();
int do_print_wrap();
int  _prints_check_data_file(char *file,int status);
int initialize_queues(char *file_name);
int str_serv(int sockfd);
int print_file(int sockfd);
int show_queue(int sockfd);
int control_queue(int sockfd);
int remove_job(int sockfd);
int send_version(int sockfd);
int get_time();
int printd_error(int sockfd);
int exec_format(char* buf, char* old_format);
int exec_gs(char* buf, char* driver, 
            char **extra_arguments, int extra_arguments_n,
            char *out_filter, char *paper_size);
int exec_add_banner(struct control *param,char* buf);
int exec_pdf_2_ps(char* buf);
int _prints_client_to_mpage(char* old_format,char* format);
int _prints_client_to_psnup(char* old_format,char* format);
                                             
int get_n_queued_jobs(int sockfd);
int readint(int fd, int *ptar);
int writen(int fd, char *ptr, int nbytes);

void printd_toggle_debug();
void printd_flag_queue_reload();
void printd_exit_cleanup();
void printd_unlink_temp_file();

int initialize_host_access(allowed_host **head);

int printmi_banner_page(FILE *fout,
                        char *user,
                        char *host,
                        char *queue,
                        char *file);

#endif /* def PrintmiPRINTDh */
