/*
  File :     printif.c


  Author :   Francesco Prelz ($Author: printmi $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
  25-Mar-1998 Created.

  Description: Simple jacket around the print client that is suitable
               for using as an lpr filter. The name of the printer is
               passed with the "af" (accounting file name, sigh)
               parameter in /etc/printcap. Unfortunately this is the
               only parameter which is passed down.
 */

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>

#include "print.h"

#define WILDCARD_UID 999999

int main (int argc, char *argv[])
 {
  int i,j,size,count;
  char temp_filename[FILENAME_MAX];
  char shell_command[512];
  FILE *input;

  struct control param;

  int sockfd;

  argc--;

  strncpy(param.printer,argv[argc],sizeof(param.printer));

  while (argv++, argc--) 
   {
    if (!strncmp(*argv, "-n", 2))
     {
      argv++,argc--;
      strncpy(param.user,*argv,sizeof(param.user));
     }

    if (!strncmp(*argv, "-h", 2))
     {
      argv++,argc--;
      strncpy(param.host,*argv,sizeof(param.host));
     }
   }

  /* Get connection to server */
  if ((is_alive(SERV_HOST_NAME_I)) == 0) /*Primary serv. down*/
   {
    if ((is_alive(SERV_HOST_NAME_II)) == 0)  /*Secondary serv. down*/
     {
      fprintf(stderr,"Cannot connect to printmi servers: %s\n",strerror(errno));
      exit(1);
     }
    sockfd=sock_init(SERV_HOST_NAME_II);
   } else {
    sockfd=sock_init(SERV_HOST_NAME_I);
   }

  input = stdin;
  strncpy(param.file,"(file from lpr)",sizeof(param.file));
  param.uid = WILDCARD_UID;
  param.reserve=0;
  param.id_only=0;
  param.copies=1;
  *param.entry='\000';

  /* The -@ flag is not used by the regular client: it will instruct
     the print server not to attempt any formatting */
  str_print(input, sockfd, "lpr", "-@", &param, TRUE);
                        
  exit(0);

 } 
